# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package sync mirrors Go package sync.

See the following link about Go sync package:

    https://golang.org/pkg/sync
"""

from __future__ import print_function, absolute_import

import threading, sys
from golang import go, defer, func, panic
from golang import context

import six

# Once allows to execute an action only once.
#
# For example:
#
#   once = Once()
#   ...
#   once.do(doSomething)
class Once(object):
    def __init__(once):
        once._mu    = threading.Lock()
        once._done  = False

    def do(once, f):
        with once._mu:
            if not once._done:
                once._done = True
                f()


# WaitGroup allows to wait for collection of tasks to finish.
class WaitGroup(object):
    def __init__(wg):
        wg._mu      = threading.Lock()
        wg._count   = 0
        wg._event   = threading.Event()

    def done(wg):
        wg.add(-1)

    def add(wg, delta):
        if delta == 0:
            return
        with wg._mu:
            wg._count += delta
            if wg._count < 0:
                panic("sync: negative WaitGroup counter")
            if wg._count == 0:
                wg._event.set()
                wg._event = threading.Event()

    def wait(wg):
        with wg._mu:
            if wg._count == 0:
                return
            event = wg._event
        event.wait()


# WorkGroup is a group of goroutines working on a common task.
#
# Use .go() to spawn goroutines, and .wait() to wait for all of them to
# complete, for example:
#
#   wg = WorkGroup(ctx)
#   wg.go(f1)
#   wg.go(f2)
#   wg.wait()
#
# Every spawned function accepts context related to the whole work and derived
# from ctx used to initialize WorkGroup, for example:
#
#   def f1(ctx):
#       ...
#
# Whenever a function returns error (raises exception), the work context is
# canceled indicating to other spawned goroutines that they have to cancel their
# work. .wait() waits for all spawned goroutines to complete and returns/raises
# error, if any, from the first failed subtask.
#
# WorkGroup is modelled after https://godoc.org/golang.org/x/sync/errgroup but
# is not equal to it.
class WorkGroup(object):

    def __init__(g, ctx):
        g._ctx, g._cancel = context.with_cancel(ctx)
        g._wg   = WaitGroup()
        g._mu   = threading.Lock()
        g._err  = None

    def go(g, f, *argv, **kw):
        g._wg.add(1)

        @func
        def _():
            defer(g._wg.done)

            try:
                f(g._ctx, *argv, **kw)
            except Exception as exc:
                with g._mu:
                    if g._err is None:
                        # this goroutine is the first failed task
                        g._err = exc
                        if six.PY2:
                            # py3 has __traceback__ automatically
                            exc.__traceback__ = sys.exc_info()[2]
                        g._cancel()
        go(_)

    def wait(g):
        g._wg.wait()
        g._cancel()
        if g._err is not None:
            # reraise the exception so that original traceback is there
            if six.PY3:
                raise g._err
            else:
                six.reraise(g._err, None, g._err.__traceback__)



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package time mirrors Go package time.

See the following link about Go time package:

    https://golang.org/pkg/time
"""

from __future__ import print_function, absolute_import

import time as _time
from golang import go, chan, select, default, nilchan, panic
import threading

# golang/py - the same as std python - represents time as float
second      = 1.0
nanosecond  = 1E-9 * second
microsecond = 1E-6 * second
millisecond = 1E-3 * second
minute      = 60   * second
hour        = 60   * minute

sleep   = _time.sleep
now     = _time.time


# ---- timers ----
# FIXME timers are implemented very inefficiently - each timer currently consumes a goroutine.

# tick returns channel connected to dt ticker.
#
# Note: there is no way to stop created ticker.
# Note: for dt <= 0, contrary to Ticker, tick returns nilchan instead of panicking.
def tick(dt):   # -> chan time
    if dt <= 0:
        return nilchan
    return Ticker(dt).c

# after returns channel connected to dt timer.
#
# Note: with after there is no way to stop/garbage-collect created timer until it fires.
def after(dt):  # -> chan time
    return Timer(dt).c

# after_func arranges to call f after dt time.
#
# The function will be called in its own goroutine.
# Returned timer can be used to cancel the call.
def after_func(dt, f):  # -> Timer
    return Timer(dt, f=f)


# Ticker arranges for time events to be sent to .c channel on dt-interval basis.
#
# If the receiver is slow, Ticker does not queue events and skips them.
# Ticking can be canceled via .stop() .
class Ticker(object):
    def __init__(self, dt):
        if dt <= 0:
            panic("ticker: dt <= 0")
        self.c      = chan(1)   # 1-buffer -- same as in Go
        self._dt    = dt
        self._mu    = threading.Lock()
        self._stop  = False
        go(self._tick)

    # stop cancels the ticker.
    #
    # It is guaranteed that ticker channel is empty after stop completes.
    def stop(self):
        with self._mu:
            self._stop = True

            # drain what _tick could have been queued already
            while len(self.c) > 0:
                self.c.recv()

    def _tick(self):
        while 1:
            # XXX adjust for accumulated error δ?
            sleep(self._dt)

            with self._mu:
                if self._stop:
                    return

                # send from under ._mu so that .stop can be sure there is no
                # ongoing send while it drains the channel.
                select(
                    default,
                    (self.c.send, now()),
                )


# Timer arranges for time event to be sent to .c channel after dt time.
#
# The timer can be stopped (.stop), or reinitialized to another time (.reset).
#
# If func f is provided - when the timer fires f is called in its own goroutine
# instead of event being sent to channel .c .
class Timer(object):
    def __init__(self, dt, f=None):
        self._f     = f
        self.c      = chan(1) if f is None else nilchan
        self._mu    = threading.Lock()
        self._dt    = None  # None - stopped, float - armed
        self._ver   = 0     # current timer was armed by n'th reset
        self.reset(dt)

    # stop cancels the timer.
    #
    # It returns:
    #
    #   False: the timer was already expired or stopped,
    #   True:  the timer was armed and canceled by this stop call.
    #
    # Note: contrary to Go version, there is no need to drain timer channel
    # after stop call - it is guaranteed that after stop the channel is empty.
    #
    # Note: similarly to Go, if Timer is used with function - it is not
    # guaranteed that after stop the function is not running - in such case
    # the caller must explicitly synchronize with that function to complete.
    def stop(self): # -> canceled
        with self._mu:
            if self._dt is None:
                canceled = False
            else:
                self._dt  = None
                self._ver += 1
                canceled = True

            # drain what _fire could have been queued already
            while len(self.c) > 0:
                self.c.recv()

            return canceled

    # reset rearms the timer.
    #
    # the timer must be either already stopped or expired.
    def reset(self, dt):
        with self._mu:
            if self._dt is not None:
                panic("Timer.reset: the timer is armed; must be stopped or expired")
            self._dt  = dt
            self._ver += 1
            go(self._fire, dt, self._ver)


    def _fire(self, dt, ver):
        sleep(dt)
        with self._mu:
            if self._ver != ver:
                return  # the timer was stopped/resetted - don't fire it
            self._dt = None

            # send under ._mu so that .stop can be sure that if it sees
            # ._dt = None, there is no ongoing .c send.
            if self._f is None:
                self.c.send(now())
                return

        # call ._f not from under ._mu not to deadlock e.g. if ._f wants to reset the timer.
        self._f()



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang import go, chan, _PanicError
from golang import sync, context
import time, threading
from pytest import raises

def test_once():
    once = sync.Once()
    l = []
    def _():
        l.append(1)

    once.do(_)
    assert l == [1]
    once.do(_)
    assert l == [1]
    once.do(_)
    assert l == [1]

    once = sync.Once()
    l = []
    def _():
        l.append(2)
        raise RuntimeError()

    with raises(RuntimeError):
        once.do(_)
    assert l == [2]
    once.do(_)  # no longer raises
    assert l == [2]
    once.do(_)  # no longer raises
    assert l == [2]


def test_waitgroup():
    wg = sync.WaitGroup()
    wg.add(2)

    ch = chan(3)
    def _():
        wg.wait()
        ch.send('a')
    for i in range(3):
        go(_)

    wg.done()
    assert len(ch) == 0
    time.sleep(0.1)
    assert len(ch) == 0
    wg.done()

    for i in range(3):
        assert ch.recv() == 'a'

    wg.add(1)
    go(_)
    time.sleep(0.1)
    assert len(ch) == 0
    wg.done()
    assert ch.recv() == 'a'

    with raises(_PanicError):
        wg.done()


def test_workgroup():
    ctx, cancel = context.with_cancel(context.background())
    mu = threading.Lock()

    # t1=ok, t2=ok
    wg = sync.WorkGroup(ctx)
    l = [0, 0]
    for i in range(2):
        def _(ctx, i):
            with mu:
                l[i] = i+1
        wg.go(_, i)
    wg.wait()
    assert l == [1, 2]

    # t1=fail, t2=ok, does not look at ctx
    wg = sync.WorkGroup(ctx)
    l = [0, 0]
    for i in range(2):
        def _(ctx, i):
            Iam__ = 0
            with mu:
                l[i] = i+1
                if i == 0:
                    raise RuntimeError('aaa')
        def f(ctx, i):
            Iam_f = 0
            _(ctx, i)

        wg.go(f, i)
    with raises(RuntimeError) as exc:
        wg.wait()
    assert exc.type       is RuntimeError
    assert exc.value.args == ('aaa',)
    assert 'Iam__' in exc.traceback[-1].locals
    assert 'Iam_f' in exc.traceback[-2].locals
    assert l == [1, 2]

    # t1=fail, t2=wait cancel, fail
    wg = sync.WorkGroup(ctx)
    l = [0, 0]
    for i in range(2):
        def _(ctx, i):
            Iam__ = 0
            with mu:
                l[i] = i+1
                if i == 0:
                    raise RuntimeError('bbb')
                if i == 1:
                    ctx.done().recv()
                    raise ValueError('ccc') # != RuntimeError
        def f(ctx, i):
            Iam_f = 0
            _(ctx, i)

        wg.go(f, i)
    with raises(RuntimeError) as exc:
        wg.wait()
    assert exc.type       is RuntimeError
    assert exc.value.args == ('bbb',)
    assert 'Iam__' in exc.traceback[-1].locals
    assert 'Iam_f' in exc.traceback[-2].locals
    assert l == [1, 2]


    # t1=ok,wait cancel  t2=ok,wait cancel
    # cancel parent
    wg = sync.WorkGroup(ctx)
    l = [0, 0]
    for i in range(2):
        def _(ctx, i):
            with mu:
                l[i] = i+1
                ctx.done().recv()
        wg.go(_, i)
    cancel()    # parent cancel - must be propagated into workgroup
    wg.wait()
    assert l == [1, 2]



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang import context, time, nilchan
from golang.context import _ready as ready
from golang.time_test import dt

# assertCtx asserts on state of _BaseCtx*
def assertCtx(ctx, children, deadline=None, err=None, done=False):
    assert isinstance(ctx, context._BaseCtx)
    assert ctx.deadline() == deadline
    assert ctx.err() is err
    assert ready(ctx.done()) == done
    assert ctx._children == children

Z = set()   # empty set
C = context.canceled
D = context.deadlineExceeded
Y = True

bg = context.background()

# test_context exercises with_cancel / with_value and merge.
# deadlines are tested in test_deadline.
def test_context():
    assert bg.err()         is None
    assert bg.done()        is nilchan
    assert bg.deadline()    is None
    assert not ready(bg.done())
    assert bg.value("hello") is None

    ctx1, cancel1 = context.with_cancel(bg)
    assert ctx1.done() is not bg.done()
    assertCtx(ctx1,     Z)

    ctx11, cancel11 = context.with_cancel(ctx1)
    assert ctx11.done() is not ctx1.done()
    assertCtx(ctx1,     {ctx11})
    assertCtx(ctx11,    Z)

    ctx111  = context.with_value(ctx11,  "hello", "alpha")
    assert ctx111.done() is ctx11.done()
    assert ctx111.value("hello") == "alpha"
    assert ctx111.value("abc")   is None
    assertCtx(ctx1,     {ctx11})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   Z)

    ctx1111 = context.with_value(ctx111, "beta",  "gamma")
    assert ctx1111.done() is ctx11.done()
    assert ctx1111.value("hello") == "alpha"
    assert ctx1111.value("beta")  == "gamma"
    assert ctx1111.value("abc")   is None
    assertCtx(ctx1,     {ctx11})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  Z)


    ctx12 = context.with_value(ctx1, "hello", "world")
    assert ctx12.done() is ctx1.done()
    assert ctx12.value("hello") == "world"
    assert ctx12.value("abc")   is None
    assertCtx(ctx1,     {ctx11, ctx12})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  Z)
    assertCtx(ctx12,    Z)

    ctx121, cancel121 = context.with_cancel(ctx12)
    assert ctx121.done() is not ctx12.done()
    assert ctx121.value("hello") == "world"
    assert ctx121.value("abc")   is None
    assertCtx(ctx1,     {ctx11, ctx12})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  Z)
    assertCtx(ctx12,    {ctx121})
    assertCtx(ctx121,   Z)

    ctx1211 = context.with_value(ctx121, "мир", "май")
    assert ctx1211.done() is ctx121.done()
    assert ctx1211.value("hello") == "world"
    assert ctx1211.value("мир")   == "май"
    assert ctx1211.value("abc")   is None
    assertCtx(ctx1,     {ctx11, ctx12})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  Z)
    assertCtx(ctx12,    {ctx121})
    assertCtx(ctx121,   {ctx1211})
    assertCtx(ctx1211,  Z)

    ctxM, cancelM = context.merge(ctx1111, ctx1211)
    assert ctxM.done() is not ctx1111.done()
    assert ctxM.done() is not ctx1211.done()
    assert ctxM.value("hello")  == "alpha"
    assert ctxM.value("мир")    == "май"
    assert ctxM.value("beta")   == "gamma"
    assert ctxM.value("abc")    is None
    assertCtx(ctx1,     {ctx11, ctx12})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  {ctxM})
    assertCtx(ctx12,    {ctx121})
    assertCtx(ctx121,   {ctx1211})
    assertCtx(ctx1211,  {ctxM})
    assertCtx(ctxM,     Z)


    for _ in range(2):
        cancel11()
        assertCtx(ctx1,     {ctx12})
        assertCtx(ctx11,    Z, err=C, done=Y)
        assertCtx(ctx111,   Z, err=C, done=Y)
        assertCtx(ctx1111,  Z, err=C, done=Y)
        assertCtx(ctx12,    {ctx121})
        assertCtx(ctx121,   {ctx1211})
        assertCtx(ctx1211,  Z)
        assertCtx(ctxM,     Z, err=C, done=Y)

    for _ in range(2):
        cancel1()
        assertCtx(ctx1,     Z, err=C, done=Y)
        assertCtx(ctx11,    Z, err=C, done=Y)
        assertCtx(ctx111,   Z, err=C, done=Y)
        assertCtx(ctx1111,  Z, err=C, done=Y)
        assertCtx(ctx12,    Z, err=C, done=Y)
        assertCtx(ctx121,   Z, err=C, done=Y)
        assertCtx(ctx1211,  Z, err=C, done=Y)
        assertCtx(ctxM,     Z, err=C, done=Y)


# test_deadline exercises deadline-related context functionality.
def test_deadline():
    t0 = time.now()
    d1 = t0 + 10*dt
    d2 = t0 + 20*dt
    d3 = t0 + 30*dt

    ctx1, cancel1 = context.with_deadline(bg, d2)
    assert ctx1.done() is not bg.done()
    assertCtx(ctx1, Z, deadline=d2)

    ctx11 = context.with_value(ctx1, "a", "b")
    assert ctx11.done() is ctx1.done()
    assert ctx11.value("a") == "b"
    assertCtx(ctx1,     {ctx11},        deadline=d2)
    assertCtx(ctx11,    Z,              deadline=d2)

    ctx111, cancel111 = context.with_cancel(ctx11)
    assert ctx111.done() is not ctx11.done
    assertCtx(ctx1,     {ctx11},        deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   Z,              deadline=d2)

    ctx1111, cancel1111 = context.with_deadline(ctx111, d3) # NOTE deadline > parent
    assert ctx1111.done() is not ctx111.done()
    assertCtx(ctx1,     {ctx11},        deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   {ctx1111},      deadline=d2)
    assertCtx(ctx1111,  Z,              deadline=d2)    # NOTE not d3

    ctx12, cancel12 = context.with_deadline(ctx1, d1)
    assert ctx12.done() is not ctx1.done()
    assertCtx(ctx1,     {ctx11, ctx12}, deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   {ctx1111},      deadline=d2)
    assertCtx(ctx1111,  Z,              deadline=d2)
    assertCtx(ctx12,    Z,              deadline=d1)

    ctxM, cancelM = context.merge(ctx1111, ctx12)
    assert ctxM.done() is not ctx1111.done()
    assert ctxM.done() is not ctx12.done()
    assert ctxM.value("a") == "b"
    assertCtx(ctx1,     {ctx11, ctx12}, deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   {ctx1111},      deadline=d2)
    assertCtx(ctx1111,  {ctxM},         deadline=d2)
    assertCtx(ctx12,    {ctxM},         deadline=d1)
    assertCtx(ctxM,     Z,              deadline=d1)

    time.sleep(11*dt)

    assertCtx(ctx1,     {ctx11},        deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   {ctx1111},      deadline=d2)
    assertCtx(ctx1111,  Z,              deadline=d2)
    assertCtx(ctx12,    Z,              deadline=d1, err=D, done=Y)
    assertCtx(ctxM,     Z,              deadline=d1, err=D, done=Y)

    # explicit cancel first -> err=canceled instead of deadlineExceeded
    for i in range(2):
        cancel1()
        assertCtx(ctx1,     Z,          deadline=d2, err=C, done=Y)
        assertCtx(ctx11,    Z,          deadline=d2, err=C, done=Y)
        assertCtx(ctx111,   Z,          deadline=d2, err=C, done=Y)
        assertCtx(ctx1111,  Z,          deadline=d2, err=C, done=Y)
        assertCtx(ctx12,    Z,          deadline=d1, err=D, done=Y)
        assertCtx(ctxM,     Z,          deadline=d1, err=D, done=Y)


    # with_timeout
    ctx, cancel = context.with_timeout(bg, 10*dt)
    assert ctx.done() is not bg.done()
    d = ctx.deadline()
    assert abs(d - (time.now() + 10*dt)) < 1*dt
    assertCtx(ctx,  Z,  deadline=d)

    time.sleep(11*dt)
    assertCtx(ctx,  Z,  deadline=d, err=D, done=Y)



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package context mirrors Go package context.

See the following links about Go contexts:

    https://blog.golang.org/context
    https://golang.org/pkg/context
"""

from __future__ import print_function, absolute_import

from golang import go, chan, select, default, nilchan
from golang import time
import threading

# Context is the interface that every context must implement.
#
# A context carries deadline, cancellation signal and immutable context-local
# key -> value dict.
class Context(object):
    # deadline() returns context deadline or None, if there is no deadline.
    def deadline(ctx):  # -> time | None
        raise NotImplementedError()

    # done returns channel that is closed when the context is canceled.
    def done(ctx):  # -> chan
        raise NotImplementedError()

    # err returns None if done is not yet closed, or error that explains why context was canceled.
    def err(ctx):   # -> error
        raise NotImplementedError()

    # value returns value associated with key, or None, if context has no key.
    def value(ctx, key):    # -> value | None
        raise NotImplementedError()


# background returns empty context that is never canceled.
def background():   # -> Context
    return  _background

# canceled is the error returned by Context.err when context is canceled.
canceled = RuntimeError("context canceled")

# deadlineExceeded is the error returned by Context.err when time goes past context's deadline.
deadlineExceeded = RuntimeError("deadline exceeded")


# with_cancel creates new context that can be canceled on its own.
#
# Returned context inherits from parent and in particular is canceled when
# parent is done.
#
# The caller should explicitly call cancel to release context resources as soon
# the context is no longer needed.
def with_cancel(parent): # -> ctx, cancel
    ctx = _CancelCtx(parent)
    return ctx, lambda: ctx._cancel(canceled)

# with_value creates new context with key=value.
#
# Returned context inherits from parent and in particular has all other
# (key, value) pairs provided by parent.
def with_value(parent, key, value): # -> ctx
    return _ValueCtx({key: value}, parent)

# with_deadline creates new context with deadline.
#
# The deadline of created context is the earliest of provided deadline or
# deadline of parent. Created context will be canceled when time goes past
# context deadline or cancel called, whichever happens first.
#
# The caller should explicitly call cancel to release context resources as soon
# the context is no longer needed.
def with_deadline(parent, deadline): # -> ctx, cancel
    # parent's deadline is before deadline -> just use parent
    pdead = parent.deadline()
    if pdead is not None and pdead <= deadline:
        return with_cancel(parent)

    # timeout <= 0   -> already canceled
    timeout = deadline - time.now()
    if timeout <= 0:
        ctx, cancel = with_cancel(parent)
        cancel()
        return ctx, cancel

    ctx = _TimeoutCtx(timeout, deadline, parent)
    return ctx, lambda: ctx._cancel(canceled)

# with_timeout creates new context with timeout.
#
# it is shorthand for with_deadline(parent, now+timeout).
def with_timeout(parent, timeout): # -> ctx, cancel
    return with_deadline(parent, time.now() + timeout)

# merge merges 2 contexts into 1.
#
# The result context:
#
#   - is done when parent1 or parent2 is done, or cancel called, whichever happens first,
#   - has deadline = min(parent1.Deadline, parent2.Deadline),
#   - has associated values merged from parent1 and parent2, with parent1 taking precedence.
#
# Canceling this context releases resources associated with it, so code should
# call cancel as soon as the operations running in this Context complete.
#
# Note: on Go side merge is not part of stdlib context and is provided by
# https://godoc.org/lab.nexedi.com/kirr/go123/xcontext#hdr-Merging_contexts
def merge(parent1, parent2):    # -> ctx, cancel
    ctx = _CancelCtx(parent1, parent2)
    return ctx, lambda: ctx._cancel(canceled)

# --------

# _Background implements root context that is never canceled.
class _Background(object):
    def done(bg):
        return nilchan

    def err(bg):
        return None

    def value(bg, key):
        return None

    def deadline(bg):
        return None

_background = _Background()

# _BaseCtx is the common base for Contexts implemented in this package.
class _BaseCtx(object):
    def __init__(ctx, done, *parentv):
        # parents of this context - either _BaseCtx* or generic Context.
        # does not change after setup.
        ctx._parentv    = parentv

        ctx._mu         = threading.Lock()
        ctx._children   = set() # children of this context - we propagate cancel there (all _BaseCtx)
        ctx._err        = None

        # chan: if context can be canceled on its own
        # None: if context can not be canceled on its own
        ctx._done       = done
        if done is None:
            assert len(parentv) == 1

        ctx._propagateCancel()

    def done(ctx):
        if ctx._done is not None:
            return ctx._done
        return ctx._parentv[0].done()

    def err(ctx):
        with ctx._mu:
            return ctx._err

    # value returns value for key from one of its parents.
    # this behaviour is inherited by all contexts except _ValueCtx who amends it.
    def value(ctx, key):
        for parent in ctx._parentv:
            v = parent.value(key)
            if v is not None:
                return v
        return None

    # deadline returns the earliest deadline of parents.
    # this behaviour is inherited by all contexts except _TimeoutCtx who overrides it.
    def deadline(ctx):
        d = None
        for parent in ctx._parentv:
            pd = parent.deadline()
            if d is None or (pd is not None and pd < d):
                d = pd
        return d

    # _cancel cancels ctx and its children.
    def _cancel(ctx, err):
        return ctx._cancelFrom(None, err)

    # _cancelFrom cancels ctx and its children.
    # if cancelFrom != None it indicates which ctx parent cancellation was the cause for ctx cancel.
    def _cancelFrom(ctx, cancelFrom, err):
        with ctx._mu:
            if ctx._err is not None:
                return  # already canceled

            ctx._err = err
            children = ctx._children;   ctx._children = set()

        if ctx._done is not None:
            ctx._done.close()

        # no longer need to propagate cancel from parent after we are canceled
        for parent in ctx._parentv:
            if parent is cancelFrom:
                continue
            if isinstance(parent, _BaseCtx):
                with parent._mu:
                    if ctx in parent._children:
                        parent._children.remove(ctx)

        # propagate cancel to children
        for child in children:
            child._cancelFrom(ctx, err)


    # propagateCancel establishes setup so that whenever a parent is canceled,
    # ctx and its children are canceled too.
    def _propagateCancel(ctx):
        pforeignv = [] # parents with !nilchan .done() for foreign contexts
        for parent in ctx._parentv:
            # if parent can never be canceled (e.g. it is background) - we
            # don't need to propagate cancel from it.
            pdone = parent.done()
            if pdone is nilchan:
                continue

            # parent is cancellable - glue to propagate cancel from it to us
            if isinstance(parent, _BaseCtx):
                with parent._mu:
                    if parent._err is not None:
                        ctx._cancel(parent._err)
                    else:
                        parent._children.add(ctx)
            else:
                if _ready(pdone):
                    ctx._cancel(parent.err())
                else:
                    pforeignv.append(parent)

        if len(pforeignv) == 0:
            return

        # there are some foreign contexts to propagate cancel from
        def _():
            _, _rx = select(
                ctx._done.recv,                         # 0
                *[_.done().recv for _ in pforeignv]     # 1 + ...
            )
            # 0. nothing - already canceled
            if _ > 0:
                ctx._cancel(pforeignv[_-1].err())
        go(_)


# _CancelCtx is context that can be canceled.
class _CancelCtx(_BaseCtx):
    def __init__(ctx, *parentv):
        super(_CancelCtx, ctx).__init__(chan(), *parentv)


# _ValueCtx is context that carries key -> value.
class _ValueCtx(_BaseCtx):
    def __init__(ctx, kv, parent):
        super(_ValueCtx, ctx).__init__(None, parent)

        # {} (key, value) specific to this context.
        # the rest of the keys are inherited from parents.
        # does not change after setup.
        ctx._kv         = kv

    def value(ctx, key):
        v = ctx._kv.get(key)
        if v is not None:
            return v
        return super(_ValueCtx, ctx).value(key)


# _TimeoutCtx is context that is canceled on timeout.
class _TimeoutCtx(_CancelCtx):
    def __init__(ctx, timeout, deadline, parent):
        super(_TimeoutCtx, ctx).__init__(parent)
        assert timeout > 0
        ctx._deadline = deadline
        ctx._timer    = time.after_func(timeout, lambda: ctx._cancel(deadlineExceeded))

    def deadline(ctx):
        return ctx._deadline

    # cancel -> stop timer
    def _cancelFrom(ctx, cancelFrom, err):
        super(_TimeoutCtx, ctx)._cancelFrom(cancelFrom, err)
        ctx._timer.stop()



# _ready returns whether channel ch is ready.
def _ready(ch):
    _, _rx = select(
            ch.recv,    # 0
            default,    # 1
    )
    if _ == 0:
        return True
    if _ == 1:
        return False
